# -*- coding: utf-8 -*-
#
# Copyright (C) 2006-2007 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://posterity.edgewall.org/wiki/License.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://posterity.edgewall.org/log/.

import re
import codecs
import email
import email.Errors
from email.Iterators import typed_subpart_iterator
import email.Utils
from datetime import datetime

from elixir import *
from sqlalchemy import object_session, relation

from posterity.util import *

SCHEMA_VERSION = 1

def init_db():
    """Initializes an empty database.
    
    This function creates and initializes the tables needed by posterity.
    """
    create_all()
    System(name='schema_version', value=str(SCHEMA_VERSION))


class UserSession(Entity):
    """The `UserSession` keeps track of the currently logged in user.
    
    """
    has_field('cookie', Unicode(32), primary_key=True)
    has_field('date', DateTime, nullable=False,
              default=datetime.utcnow)
    belongs_to('account', of_kind='Account', required=True)
    using_options(tablename='user_session')
    
    def __init__(self):
        super(UserSession, self).__init__()
        self.cookie = hex_entropy()
    
    def logout(self):
        for us in UserSession.select_by(account_id=self.account.id):
            objectstore.delete(us)


class System(Entity):
    """Table for storage of system wide settings.

    This table is for example used to keep track of the SQL schema version
    (name='schema_version')
    """
    has_field('name', Unicode(100), primary_key=True)
    has_field('value', Unicode)
    using_options(tablename='system')


class Account(Entity):
    
    has_field('id', Integer, primary_key=True)
    has_field('username', Unicode(100), nullable=False, unique=True)
    has_field('password', Unicode, nullable=False)
    has_many('tags', of_kind='Tag', order_by='name')
    has_many('identities', of_kind='Identity', order_by='name')
    using_options(tablename='account')

    @staticmethod
    def create(username):
        """Creates a new account and the required tags"""
        a = Account(username=username, password='')
        Tag(account=a, name='All Mail', is_special=True)
        Tag(account=a, name='Inbox', is_special=True)
        Tag(account=a, name='Sent Mail', is_special=True)
        Tag(account=a, name='Spam', is_special=True)
        Tag(account=a, name='Trash', is_special=True)
        return a

    def set_password(self, password):
        self.password = encrypt_password(password)


class Identity(Entity):
    """Email sending identity
    
    When an email is sent using posterity the name, email and signature
    is taken from the identity. Each account can be configured with 
    zero or more identities.
    """
    has_field('id', Integer, primary_key=True)
    has_field('name', Unicode, nullable=False)
    has_field('email', Unicode, nullable=False)
    has_field('signature', Unicode, nullable=False)
    has_field('reply_to', Unicode, nullable=False)
    has_field('organization', Unicode, nullable=False)
    has_field('is_default', Boolean, nullable=False, default=False)
    belongs_to('account', of_kind='Account', required=True)
    using_options(tablename='identity')


class Msg(Entity):
    """Each `Msg` instance represents an email message

    The actual RFC2822 data is stored in the payload column unchanged.
    But for performance reasons a couple of important header values
    are extracted and stored in dedicated columns for easy access.
    """
    has_field('id', Integer, primary_key=True)
    has_field('is_deleted', Boolean, nullable=False, default=False)
    has_field('is_unread', Boolean, nullable=False, default=True)
    has_field('msg_id', Unicode(200))
    has_field('payload', BLOB, nullable=False, deferred=True)
    has_field('subject', Unicode, nullable=True)
    has_field('from_', Unicode, nullable=True)
    has_field('from_name', Unicode, nullable=True)
    has_field('from_email', Unicode, nullable=True)
    has_field('to_', Unicode, nullable=True, key='to')
    has_field('cc', Unicode, nullable=True)
    has_field('date', DateTime, nullable=False, default=datetime.utcnow)
    belongs_to('account', of_kind='Account', required=True)
    belongs_to('conv', of_kind='Conv', required=True)
    using_options(tablename='msg')

    @staticmethod
    def parse_message(account, message):
        """Parses an `email.Message.Message`

        Figure out which conversation the message belongs to and stores 
        it in the database.
        """
        conv = Conv.for_message(account, message)
        msg_id = message['Message-ID']
        subject = decode_header(message['Subject'])
        from_name, from_email = email.Utils.parseaddr(str(message['From']))
        from_name = decode_header(from_name)
        from_email = decode_header(from_email)
        if from_name:
            from_ = '%s <%s>' % (from_name, from_email)
        else:
            from_ = from_email
        to = decode_header(message['To'])
        cc = decode_header(message['Cc'])
        conv.date = datetime.utcnow()

        msg = Msg(account=account, conv=conv, subject=subject, msg_id=msg_id,
                  payload=message.as_string(), to=to, cc=cc,
                  from_name=from_name, from_email=from_email, from_=from_)
        conv._update_date_cache()
        return msg

    def quote_body(self):
        """Returns a quoted version of the message.

        The first line will begin with 'X wrote:'
        """
        data = []
        message = email.message_from_string(self.payload)
        for part in typed_subpart_iterator(message, 'text', 'plain'):
            data.append(decode_text_plain_part(part))

        return '%s wrote:\r\n%s' % (self.from_name or self.from_email,
                                  quote_text('\n'.join(data)))


class Conv(Entity):
    """
    A Conversation is a set of related messages. 
    
    All messages usually belong to the same thread of conversation.
    """
    has_field('id',  Integer, primary_key=True)
    has_and_belongs_to_many('tags', of_kind='Tag', tablename='conv_to_tag', 
                            lazy=False)
    belongs_to('account', of_kind='Account', required=True)
    using_options(tablename='conv')

    @staticmethod
    def for_message(account, message):
        """Returns the `Conv` instance to which `message` belongs.
        
        If no existing conversation is found a new is created.
        The 'in-reply-to' and 'references' headers if present are used
        to locate the conversation. If these headers are missing the subject
        header is used as a last resort.
        """
        inreplyto = message.get('in-reply-to',
                                message.get('references'))
        match = inreplyto and re.search('<[^>]+>', inreplyto)
        if match:
            c = Conv.selectfirst(and_(Msg.c.msg_id==inreplyto,
                                   Msg.c.conv_id==Conv.c.id,
                                   Conv.c.account_id==account.id))
            if c:
                return c

        if message['Thread-Topic']: # Non-standard header used by MS exchange
            # If no conversation is found so far we need to fall back to
            # finding conversations with a similar subject
            subject = decode_header(message['Thread-Topic'])
            c = Conv.selectfirst(and_(Msg.c.subject==subject,
                                   Msg.c.conv_id==Conv.c.id,
                                   Conv.c.account_id==account.id),
                              order_by=desc(Msg.c.date))
            if c:
                return c
        return Conv(account=account, subject=message['subject'])

    def add_tag(self, tag):
        """Associate the current conversation with `tag`"""
        if isinstance(tag, basestring):
            for t in self.account.tags:
                if t.name == tag:
                    self.tags.append(t)
                    break
            else:
                raise ValueError
        else:
            self.tags.append(tag)
        self._update_date_cache()

    def remove_tag(self, tag):
        """Untag the current conversation"""
        if isinstance(tag, basestring):
            for t in self.account.tags:
                if t.name == tag:
                    tag = t
                    break
            else:
                raise ValueError
        if tag in self.tags:
            self.tags.remove(tag)
            self._update_date_cache()

    def mark_as_deleted(self):
        """Mark the entire conversation and all associated messages as deleted.

        Deleted conversations and messages are not normally displayed in a
        conversation index or conversation view. The 'Trash' and 'Spam' tags
        are exceptions to this rule.
        """
        for msg in self.msgs:
            msg.is_deleted = True
        self._update_date_cache()

    def undelete(self):
        """Mark the entire conversation and all associated messages as deleted.

        Deleted conversations and messages are not normally displayed in a
        conversation index or conversation view. The 'Trash' and 'Spam' tags
        are exceptions to this rule.
        """
        for msg in self.deleted_msgs:
            msg.is_deleted = False
        self._update_date_cache()

    def mark_as_read(self, deleted=False):
        """Mark all messages in the conversation as read
        
        If @deleted is True all messages marked for deletions are marked
        as read.
        """
        if deleted:
            for msg in self.deleted_msgs:
                msg.is_unread = False
        else:
            for msg in self.msgs:
                msg.is_unread = False
            
    def _update_date_cache(self):
        """
        In order to speed up the conversation index query the maximum 
        message timestamp for each conversation is stored directly in
        the column conv_to_tag.c.conv_date.
        
        The conv_to_tag.c.conv_date2 column contains the corresponding
        timestamp for "deleted" messages.
        """
        conn = local.trans.connection(Conv)
        objectstore.flush()
        s = select([func.max(Msg.c.date)], 
                   and_(Msg.c.conv_id==self.id,
                        Msg.c.is_deleted==False))
        s = conv_to_tag.update(and_(conv_to_tag.c.conv_id==self.id),
                               values={conv_to_tag.c.conv_date: s})
        conn.execute(s)
        s = select([func.max(Msg.c.date)], 
                   and_(Msg.c.conv_id==self.id,
                        Msg.c.is_deleted==True))
        s = conv_to_tag.update(and_(conv_to_tag.c.conv_id==self.id),
                               values={conv_to_tag.c.conv_date2: s})
        conn.execute(s)

    @staticmethod
    def search(query, limit=100, offset=0, deleted=False):
        """Return conversations matching the `query`"""
        convs = Conv.select(and_(Conv.c.id==Msg.c.conv_id,
                                 Msg.c.is_deleted==deleted,
                                 Msg.c.subject.like('%'+query+'%')),
                            order_by=desc(Msg.c.date),
                            offset=offset, limit=limit, distinct=True)
        return convs

Conv.mapper.add_property('msgs', 
                         relation(Msg, lazy=False,
                                  primaryjoin=and_(Conv.c.id==Msg.c.conv_id,
                                                   not_(Msg.c.is_deleted))))
Conv.mapper.add_property('deleted_msgs', 
                         relation(Msg, lazy=False,
                                  primaryjoin=and_(Conv.c.id==Msg.c.conv_id,
                                                   Msg.c.is_deleted)))


class Tag(Entity):

    has_field('id', Integer, primary_key=True)
    has_field('name', Unicode(100), nullable=False)
    has_field('is_special', Boolean, nullable=False, default=False)
    belongs_to('account', of_kind='Account', required=True)
    has_and_belongs_to_many('convs', of_kind='Conv', tablename='conv_to_tag')
    using_options(tablename='tag')

    def get_convs(self, deleted=False, offset=0, limit=100):
        if deleted:
            convs = Conv.filter(and_(conv_to_tag.c.conv_id==Conv.c.id,
                                     conv_to_tag.c.tag_id==self.id,
                                     conv_to_tag.c.conv_date2 != None,
                                     Conv.c.account_id==self.account_id))
            convs = convs.order_by(desc(conv_to_tag.c.conv_date2))
        else:
            convs = Conv.filter(and_(conv_to_tag.c.conv_id==Conv.c.id,
                                     conv_to_tag.c.tag_id==self.id,
                                     conv_to_tag.c.conv_date != None,
                                     Conv.c.account_id==self.account_id))
            convs = convs.order_by(desc(conv_to_tag.c.conv_date))
        return convs.offset(offset).limit(limit).list()

# We need to add these two columns manually since this table is created
# implicitly by sqlalchemy
conv_to_tag = metadata.tables['conv_to_tag']
conv_to_tag.append_column(Column('conv_date', DateTime))
conv_to_tag.append_column(Column('conv_date2', DateTime))

Index('tag_name_unique_idx', Tag.c.account_id, Tag.c.name, unique=True)
Index('msg_msg_id_idx', Msg.c.msg_id)
Index('conv_to_tag_unique_idx', conv_to_tag.c.tag_id, conv_to_tag.c.conv_id,
      unique=True)
Index('conv_to_tag_conv_date_idx', conv_to_tag.c.conv_date)
Index('conv_to_tag_conv_date2_idx', conv_to_tag.c.conv_date2)

